# help.tcl -- Generalized help system for tcl/tk scripts.
# Copyright (c) 1995 by James F. Carter.  950304

# Show a help item.  Arguments:
#	wid	Name that the help widget should give itself.  The help widget
#		(a toplevel window) will be positioned relative to its parent.
#	key	Which item in the help file should be shown.  {} means to open
#		the help browser immediately.
#	filename Full name with directory of the help file, or a relative path
#		name, which will be relative to the last help file used.

# Global var $help contains all the internal data for the help system.
# help(filename_key) = {start length category condition description}
#	start		Seek pointer to start of item text
#	length		Length of text in bytes
#	category	Category code for sub-index.  Specify ALL (upper case)
#			if you don't have sub-indices.
#	condition	For games, where the user isn't allowed to see some
#			help.  Specify as 1 if all help is visible all the time.
#	description	Descriptive text, may contain blanks.
# A help file consists of concatenated pieces each beginning with a line (not
# shown to the user) of "==== key category condition description", whitespace
# separated, content as above.  3 or more = are required as the first bytes
# on a line.  

# Within the body text, the sequence {key text} is a hyperlink.
# "key" is the key (as in the header) of the item to show when the link
# is clicked on.  "text" (minus {}, key and one blank) is shown, highlighted,
# and clickable.  "text" may contain interior whitespace (but not at the ends)
# and may cross line boundaries (but not paragraphs).
# If the "condition" for the key is not met at the time of displaying, the
# entire hyperlink is omitted.  Special case, { text } (with at least one
# blank before and after the text) is not recognized as a hyperlink.

# Newline-whitespace (including newline-newline) counts as a paragraph 
# boundary and is preserved as-is.  Within a paragraph, existing newlines
# are changed to blanks and words are wrapped to fit.  (A word wider than the 
# text widget will be broken in the middle.)  Lines after the first begin at 
# the left edge of the text widget -- no dynamic margins.  Internal whitespace
# is preserved, except that one blank at the end of a line will be removed.  
# If you want two blanks after periods, put two blanks after any period at 
# the end of a line.

set helpcats(ALL) {}	;#Lists of keys in each category
set helpcats(HISTORY) {} ;#Last N topics looked at, most recent last.
set help(BACKHIST) {}	;#{key prior}: how we got to the last N items.
set help(NEXTPAGE) {}	;#fkey of the page about to be read.
set helpok(1) 1		;#Help conditions to lock people out of certain info
proc help {wid key filename} {
    global help helpcats helpok helpfiles
		#Set up the internal key: filename#key.
    if {! [info exists help(DIR_NAME)]} {
	set help(DIR_NAME) [file dirname $filename]
    }
    set filename [file tail $filename]
    set fkey ${filename}#$key
		#Create the help viewer, if not done yet.
    help_new $wid $fkey
		#Show the info.  If it can't be shown, put up a browser instead.
    if {$key == {} || ! [help_show $wid $fkey {}]} {
	help_browser $wid $fkey {}
    }
}

# Prepare the help viewer for use.  Create it if not created yet.  $fkey = 
# filename#(ignored) for the help browser on the history and search button.
proc help_new {wid fkey} {
    set wit $wid.text
    if {! [winfo exists $wid]} {
	regsub {\.[^.]*$} $wid {} parent
	if {$parent == {}} {set parent .}
	toplevel $wid
	wm geometry $wid +[expr [winfo rootx $parent]+15]+[expr [winfo rooty $parent]+15]
	wm title $wid "[wm title .] help"
	frame $wid.fr
	pack  $wid.fr -in $wid -fill x -expand 1
	set j 0
	foreach cmd "\{help_browser $wid $fkey \{\}\}\
		     \{help_back $wid\}\
		     \{help_browser $wid $fkey HISTORY\}\
		     \{destroy $wid\}" {
	    set bt [lindex {Search Back History Close} $j]
	    set b [string tolower $bt]
	    button $wid.fr.$b -text $bt -command $cmd -width 8
	    pack $wid.fr.$b -in $wid.fr -side left
	    incr j
	}
	label $wid.fr.lbl -relief raised
	pack  $wid.fr.lbl -in $wid.fr -fill x -expand 1
	scrollbar $wid.textsb -command "$wit yview" -relief sunken
	pack $wid.textsb -in $wid -side left -fill y
	text $wit -relief sunken -yscrollcommand "$wid.textsb set" 
	pack $wit -in $wid -side top -padx 2 -pady 2
	update			;#Ensure the viewer is realized, for browser
    }
    $wid.fr.lbl configure -text "          "
    return $wid
}

# Ensure that the header for a particular key is loaded.  If this is the first
# time the key was seen, the data will not be read until later, and so the
# data length will be missing (zero).  If fkey is filename# (nothing) then
# it will never be found and the entire file will be read.  This is what you
# do when opening the browser.  A synthetic key of filename# will be generated.
# Return value is 1 if there is a help page that can be read, 0 if not.
proc help_readheader {fkey} {
    global help helpcats
    regexp {^([^#]*)#(.*)$} $fkey junk filename key
    if {$key != {} && [info exists help($fkey)]} {return 1}	;#Yes it's here
    if {[info exists help(${filename}#)]} {return 0}	;#EOF read, no such key
		#Get the help file open.
    if {! [info exists help($filename)]} {
	set help($filename) [open $help(DIR_NAME)/$filename]
    }
    set fd $help($filename)
    set caught $help(NEXTPAGE)		;#Key of last header read.
    set eof 0
		#If help(NEXTPAGE) is nonnull then the file was left after 
		#that header.  If null, it was left before a header, possibly
		#several lines before.
    set hafter [tell $fd]
    while 1 {
			#If the line about to be read were a header, hbefore
			#would be the end of the previous page, and hafter would
			#be the beginning of the next one.
	set hbefore $hafter
	if {[gets $fd line] < 0} {
	    set eof 1
	    set line "===== {} EOF 1 EOF"	;#Synthetic key symbolizing EOF.
	}
	set hafter [tell $fd]
		    #Got an item header line
	if {[string match "===*" $line]} {
		    #Fill in the length of the last item.
	    if {$caught != {}} {
		set help($caught) [lreplace $help($caught) 1 1 [expr $hbefore-[lindex $help($caught) 0]]]
	    }
		    #Save index item about the item that is beginning.
	    set caught ${filename}#[lindex $line 1]
	    set help($caught) "$hafter 0 [lrange $line 2 end]"
	    set foundit [expr {$caught == $fkey}]
	    set cat [lindex $line 2]
	    if {$cat != "EOF"} {
		lappend helpcats($cat) $caught
		if {$cat != "ALL"} {lappend helpcats(ALL) $caught}
	    } 
	    if {$foundit || $eof} break
	}
    }
    set help(NEXTPAGE) $fkey		;#Key for page about to be read
    return $foundit
}

#Copy the item into the text widget, wrapping words.  This is NOT a public
#entry point.  Arguments:
#	wid	Name of widget
#	fkey	filename#key that will be shown.  It may not be {} but the
#		program is idiotproof if it doesn't exist or is secret.
#	prior	filename#key of frame containing hyperlink to $fkey (or {})
# Return value	1 if something was displayed; 0 if it wasn't.
proc help_show {wid fkey prior} {
    global help helpcats
			#Preliminary setup of the viewer widget
    wm deiconify $wid
    raise $wid
    set lbl $wid.fr.lbl
    set wit $wid.text
    $wit configure -state normal
    $wit delete 1.0 end				;#Erase all previous text.
			#Keep record of the last 30 (unique) things looked at, 
			#most recent first.  Including bogus targets, so the
			#user can get back via the back key.
    set j [lsearch $helpcats(HISTORY) $fkey]
    if {$j < 0 && [llength $helpcats(HISTORY)] >= 30} {set j 29}
    if {$j >= 0} {set helpcats(HISTORY) [lreplace $helpcats(HISTORY) $j $j]}
    set helpcats(HISTORY) [concat $fkey $helpcats(HISTORY)]
			#Keep a record of the most recent route by which we got
			#to this help item.  This is used by the "Back" key.
    if {$prior != {}} {
	set j [lsearch -glob $help(BACKHIST) "$fkey *"]
	if {$j >= 0} {
	    set help(BACKHIST) [lreplace $help(BACKHIST) $j $j "$fkey $prior"]
	} else {
	    set help(BACKHIST) [linsert [lrange $help(BACKHIST) 0 28] 0 "$fkey $prior"]
	}
    }
    set help(CURKEY) $fkey
			#Deal with special cases that prevent displaying.
    if {! [help_readheader $fkey]} {
	$lbl configure -text "$fkey not found"
	return 0
    }
    if {! [help_secret $fkey]} {
	$lbl configure -text "$fkey is restricted information"
	return 0
    }
    set hdr $help($fkey)
		#Widget initialization for showing the help page
    $wid.fr.lbl configure -text [lrange $hdr 4 end]
    set width [lindex [$wit configure -width] 4]
    incr width -1	;#Compensate for 4.0b3 bug: leave 1 blank at end
    regexp {^([^#]*)#(.*)$} $fkey junk filename key
    set fd $help($filename)
		#If the length is known, read the data all at once.
		#If not known yet, read it line by line and stop at the header.
    if {[lindex $hdr 1] <= 0} {
	set data {}
	set line {}
	set here [lindex $hdr 0]
	while {[gets $fd line] >= 0} {			;#While not EOF
	    if {[string match "===*" $line]} {break}	;#Hit next page header
	    append data "$line\n"
	    set here [tell $fd]
	}
			#Save the actual length of the data.
	set help($fkey) [lreplace $hdr 1 1 [expr $here-[lindex $hdr 0]]]
	seek $fd $here		;#Back up so header or EOF is read again.
	set help(NEXTPAGE) {}	;#Signifying it's not after a header.
    } else {
	set here [tell $fd]
	seek $fd [lindex $hdr 0]
	set data [read $fd [lindex $hdr 1]]
	seek $fd $here		;#Leave file position unchanged.
    }
		#Within a paragraph, convert newlines to spaces.  Newline-
		#whitespace or newline-newline are paragraph boundaries and
		#are retained.  Avoid creating double blanks at line ends.
    regsub -all "(\[^\n\]) ?\n(\[^ \t\n\])" $data {\1 \2} data
		#Prepare the hypertext links.  They're read as {key text};
		#convert to {+nbr text }-nbr (with spaces exactly as noted),
		#except delete any that are not supposed to be seen.
		# $tag2key(nbr) gives the original key, and $tagst(nbr) will 
		#give the text widget index where the link's text starts.
		#Note, { stuff } (with those spaces) is ignored as a hyperlink.
    set tagn 0			;#Tag number for hypertext links
    set data2 {}
    foreach sect [split $data "\{"] {   ;#\}
		#Grab the key (1st section, and sect. beginning with a blank,
		#are not hyperlinks)
	set linked 0
	if {$tagn > 0 && [regexp "^\[^ \]+" $sect hkey]} {
	    set lhkey [string length $hkey]
	    if {[string first "#" $hkey] < 0} {set hkey "${filename}#$hkey"};#\{
	    if {[set j [string first "\}" $sect]] > 0 && \
			[help_readheader $hkey]} {     ;#If has a referent
		if {[help_secret $hkey]} {	       ;#If user may see it
		    set tag2key($tagn) $hkey
		    append data2 "\{$tagn [string range $sect [expr $lhkey+1] [expr $j-1]]\}$tagn "
		}
		set sect [string range $sect [incr j] end]
		set linked 1
	    }
	}
	if {! $linked && $tagn > 0} {
	    append data2 "\{"	;#\}
	    if {! [regexp {^ } $sect]} {append data2 " "}
	}
	append data2 $sect
	incr tagn
    }
		#Digest each paragraph independently.
    set linen 1			;#Line number, used in hypertext links
    foreach para [split $data2 "\n"] {
	set para "[string trimright $para] "	;#End with exactly 1 blank.
	while 1 {
			#Locate the last blank before (width) bytes.
	    if {! [regexp -indices { +[^ ]*$} [string range $para 0 $width] jj]} {
		if {$para == {}} break
		set jj "$width $width"		;#very long word, break at width
	    }
	    set j [lindex $jj 0]
	    set line "[string range $para 0 [expr $j-1]]\n"
	    set para [string trimleft [string range $para $j end]]
			#Insert pieces of the line in the text widget, splitting
			#on hyperlink boundaries.  Remember beginnings; set
			#the tag on endings.
	    regsub -all {[{}]} $line "\a&" line
	    foreach sect [split $line "\a"] {
		set linked [regexp {^([{}])([0-9]+)} $sect junk end tagn]
		if $linked {
		    if {$end == "\{"} {		;#\}
			set tagst($tagn) [$wit index insert]
		    } elseif {[info exists tagst($tagn)]} {
			set tagend [$wit index insert]
			set linked 2
		    }
		    regsub {^.[0-9]+ ?} $sect {} sect
		}
		$wit insert insert $sect
		if {$linked == 2} {
		    $wit tag add tag$tagn $tagst($tagn) $tagend
		    if {[regexp {color} [winfo visual .]]} {
				#Color Spring Green 2
			$wit tag configure tag$tagn -background rgb:00/ee/76
		    } else {
			$wit tag configure tag$tagn -underline 1
		    }
		    $wit tag bind tag$tagn <Button-1> \
			"help_show $wid $tag2key($tagn) $fkey"
		}
	    }
	}
    }
    $wit yview 0
    $wit configure -state disabled
    return 1
}


# Check a condition.  Argument is the key for the index entry.
# Return value is 1 if user may see the entry, 0 if not.
proc help_secret {key} {
    global help helpok
#    set debug "help_secret($key): exists [info exists help($key)]"
#    if {[info exists help($key)]} {
#	set cond [lindex $help($key) 3]
#	append debug " cond=$cond"
#	if {[info exists helpok($cond)]} {
#	    append debug " helpok(cond) = $helpok($cond)"
#	} else {
#	     append debug " not in helpok"
#	}
#    }
#    puts $debug
    if {! [info exists help($key)]} {return 1}
    set cond [lindex $help($key) 3]
    return [expr {[info exists helpok($cond)] && $helpok($cond) != 0}]
}

# Return to the "previous" topic.
proc help_back {wid} {
    global help
    set fkey $help(CURKEY)
    set j [lsearch -glob $help(BACKHIST) "$fkey *"]
    if {$j < 0} {
	puts "Didnt find \"$fkey *\" in $help(BACKHIST)" ;#DEBUG
	help_browser $wid $fkey HISTORY
    } else {
	help_show $wid [lindex [lindex $help(BACKHIST) $j] 1] {}
    }
}

# Help browser.  This is NOT a public entry point.  Arguments: 
#	wid	Name of main help widget
#	fkey	Key in the form filename#(possibly with stuff following).
#	cat	Category to start with, or {} to start by showing categories.
proc help_browser {wid fkey cat} {
    global tk_version
    set wbr $wid.br
    if {! [winfo exists $wbr]} {
	frame $wbr -relief raised -borderwidth 2
	place $wbr -in $wid -x 15 -y 15
	# set geom +[expr [winfo rootx $wid]+15]+[expr [winfo rooty $wid]+15]
	# wm geometry $wbr $geom
	# wm title $wbr "Help Browser"
	frame $wbr.fr
	pack  $wbr.fr -in $wbr -fill x -expand 1
	label $wbr.fr.lbl
	pack  $wbr.fr.lbl -in $wbr.fr -side right -fill x -expand 1
	foreach b {categories close} {
	    button $wbr.fr.$b -text $b
	    pack $wbr.fr.$b -in $wbr.fr -side left
	}
	$wbr.fr.categories configure -command "help_cats $wid"
	$wbr.fr.close  configure -command "help_br_destroy $wid"
	$wid.fr.close  configure -command "destroy $wbr; destroy $wid"
	set lbox $wbr.lbox
	scrollbar $wbr.boxsb -command "$lbox yview" -relief sunken
	pack $wbr.boxsb -in $wbr -side left -fill y
	set cmd [list listbox $lbox -yscrollcommand "$wbr.boxsb set" \
	    -relief sunken]
	if {[string compare $tk_version 4.0] >= 0} {
	    lappend cmd -width 40 -height 12
	} else {
	    lappend cmd -geometry 40x12
	}
	eval $cmd
	pack $lbox -in $wbr -side right
    }
    # wm deiconify $wbr
    raise $wbr
		#Scan the whole help file for headers.  Load into listbox.
    regsub {#.*$} $fkey "#" fkey
    help_readheader $fkey
    if {$cat == {}} {
	help_cats $wid
    } else {
	help_items $wid $cat
    }
}

proc help_br_destroy {wid} {
    destroy $wid.br
    $wid.fr.close  configure -command "destroy $wid"
}

# Load the categories into the list box and configure appropriately.
proc help_cats {wid} {
    global help helpcats helpok
    set wbr $wid.br
    set lbox $wbr.lbox
    set cats [lsort [array names helpcats]]
    if {$cats == "ALL HISTORY"} {
	help_items $wid ALL
    } else {
	$lbox delete 0 end
	bind $lbox <Button-1> "help_cats_b $wid %y"
		#Only load categories with at least one accessible item.
	foreach c $cats {
	    set ok 0
	    foreach ci $helpcats($c) {
		if {[help_secret $ci]} {
		    set ok 1
		    break
		}
	    }
	    if {$ok} {
		$lbox insert end $c
	    }
	}
	$lbox yview 0
	$wbr.fr.lbl configure -text Categories
    }
}

# React to a button press in the category listbox
proc help_cats_b {wid y} {
    global tk_version
    set lbox $wid.br.lbox
    set y [$lbox nearest $y]
    if {[string compare $tk_version 4.0] >= 0} {
	$lbox selection clear 0 end
	$lbox selection anchor $y
    } else {
	$lbox select clear
	$lbox select from $y
    }
    help_items $wid [$lbox get $y]
}

# Load the items from a category into the list box and configure appropriately.
proc help_items {wid cat} {
    global help helpcats helpok
    if {$cat == {}} {
	help_cats $wid
	return
    }
    set lbox $wid.br.lbox
    $lbox delete 0 end
    bind $lbox <Button-1> "help_items_b $wid $cat %y"
    set help(LOADED) {}
    foreach c $helpcats($cat) {
	set hitem $help($c)
	if {[help_secret $c]} {
	    lappend help(LOADED) $c
	    $lbox insert end [lrange $help($c) 4 end]
	}
    }
    $lbox yview 0
    $wid.br.fr.lbl configure -text $cat
}

# React to a button press in the items listbox
proc help_items_b {wid cat y} {
    global help helpcats tk_version
    set lbox $wid.br.lbox
    set y [$lbox nearest $y]
#    if {[string compare $tk_version 4.0] >= 0} {
#	$lbox selection clear 0 end
#	$lbox selection anchor $y
#    } else {
#	$lbox select clear
#	$lbox select from $y
#    }
    help_show $wid [lindex $help(LOADED) $y] {}
    help_br_destroy $wid
}
